{

  GIF file analyzer by melanion.info

}

unit GifAnalyzeClassUnit;

interface

uses
  System.SysUtils, System.Types, System.UITypes, System.Classes,
  System.Math, System.Generics.Collections;

type
  TBlockType = (btHeader, btAppExt, btCtrlExt, btImgBlock,
                btCommentExt, btTextExt, btTrailer, btUnknown);

  PHeader = ^THeader;
  THeader = record
    Signature: array[0..2] of AnsiChar;
    Version:  array[0..2] of AnsiChar;
    LogicalScreenWidth: Word;
    LogicalScreenHeight: Word;
    TheBits: Byte;
    BackgroundColorIndex: Byte;
    PixelAspectRatio: Byte;
    ColorTable: array[0..767] of byte;
    function GlobalColorTableFlag: boolean;
    function ColorResolution: integer;
    function SortFlag: boolean;
    function SizeOfGlobalColorTable: integer;
    function SizeOfColorTable: integer;
  end;

  PBlock = ^TBlock;
  TBlock = record
    BlockSize: byte;
    Data: array[0..255] of byte;
  end;

  PApplicationExtension = ^TApplicationExtension;
  TApplicationExtension = record
    Address: integer;
    Size: integer;
    ExtensionIntroducer: byte;
    ExtensionLabel: byte;
    BlockSize: byte;
    ApplicationIdentifier: array[0..7] of AnsiChar;
    ApplicationAuthenticationCode: array[0..2] of AnsiChar;
    Blocks: TList;  // <PBlock> need check
  end;

  PGraphicControlExtension = ^TGraphicControlExtension;
  TGraphicControlExtension = record
    Address: integer;
    Size: integer;
    Index: integer;
    ExtensionIntroducer: byte;   // 0x21 Fixed value
    GraphicControlLabel: byte;   // 0xf9 Fixed value
    BlockSize: byte;             // 0x04 Fixed value
    TheBits: Byte;
    DelayTime: Word;             // Frame delay (1/100 sec)
    TransparentColorIndex: byte;
    BlockTerminator: byte;       // 0
    function DisposalMothod: integer;
    function UserInputFlag: boolean;
    function TransparentColorFlag: boolean;
  end;

  PImageBlock = ^TImageBlock;
  TImageBlock = record
    Address: integer;
    Size: integer;
    Index: integer;
    ImageSeparator: byte;        // 0x2c Fixed value
    ImageLeftPosition: Word;
    ImageTopPosition: Word;
    ImageWidth: Word;
    ImageHeight: Word;
    TheBits: Byte;
    LocalColorTable: array[0..767] of byte;
    LZWMinimumCodeSide: byte;
    Blocks: TList{<PBlock>};
    // BlockTerminator: byte;       // 0
    function LocalColorTableFlag: boolean;
    function InterlaceFlag: boolean;
    function SortFlag: boolean;
    function SizeOfLocalColorTable: integer;
    function SizeOfColorTable: integer;
  end;

  PCommentExtension = ^TCommentExtension;
  TCommentExtension = record
    Address: integer;
    Size: integer;
    ExtensionIntroducer: byte;
    CommentLabel: byte;
    Blocks: TList{<PBlock>};
  end;

  PTextExtension = ^TTextExtension;
  TTextExtension = record
    Address: integer;
    Size: integer;
    ExtensionIntroducer: byte;  // $21
    PlainTextLabel: byte;       // $01
    BlockSize: byte;            // $0C
    TextGridLeftPosition: Word;
    TextGridTopPosition: Word;
    TextGridWidth: Word;
    TextGridHeight: Word;
    CharacterCellWidth: byte;
    CharacterCellHeight: byte;
    TextForegroundColorIndex: byte;
    TextBackgroundColorIndex: byte;
    Blocks: TList{<PBlock>};
  end;

  PItemRec = ^TItemRec;
  TItemRec = record
    BlockType: TBlockType;
    Caption: string;
    Info: string;
    Block: pointer;
  end;

  TAnimatedGifClass = class
  strict private
    MS: TMemoryStream;
    FBlockList: TList;
    FControlExtensionCount: integer;
    FImageBlockCount: integer;
    FNextByte: byte;
    FTrailer: byte;
    FProcessing: boolean;
    // clear
    procedure DataClear;
    procedure TextExtensionClear(Rec: PTextExtension);
    procedure CommentExtensionClear(Rec: PCommentExtension);
    procedure ImageBlockClear(Rec: PImageBlock);
    procedure ApplicationExtensionClear(Rec: PApplicationExtension);

    // Decode
    procedure DecodeNow;
    procedure DecodeHeader;
    procedure DecodeGraphicControlExtension;
    procedure DecodeImageBlock;
    procedure DecodeCommentExtension;
    procedure DecodePlainTextExtension;
    procedure DecodeApplicationExtension;
    procedure DecodeUnknownExtension;

    // Encode
    function EncodeNow: boolean;
    procedure WriteData(Rec: PBlock);
    procedure EncodeHeader(Rec: PHeader);
    procedure EncodeGraphicControlExtension(Rec: PGraphicControlExtension);
    procedure EncodeImageBlock(Rec: PImageBlock);
    procedure EncodeCommentExtension(Rec: PCommentExtension);
    procedure EncodePlainTextExtension(Rec: PTextExtension);
    procedure EncodeApplicationExtension(Rec: PApplicationExtension);

    function GetBlockListCount: integer;
    function GetBlockListRec(index: integer): PItemRec;
  public
    property BlockList[index: integer]: PItemRec read GetBlockListRec;
    property BlockCount: integer read GetBlockListCount;
    property ImageBlockCount: integer read FImageBlockCount;
    property Trailer: byte read FTrailer;
    constructor Create;
    destructor Destroy; override;
    procedure LoadData(Filename: string);
    procedure SaveData(Filename: string);
    // Gif Info
    function GetHeaderInfo(Rec: PHeader): string;
    function GetGraphicControlExtensionInfo(Rec: PGraphicControlExtension): string;
    function GetImageBlockInfo(Rec: PImageBlock): string;
    function GetCommentExtensionInfo(Rec: PCommentExtension): string;
    function GetPlainTextExtensionInfo(Rec: PTextExtension): string;
    function GetApplicationExtensionInfo(Rec: PApplicationExtension): string;
    function GetUnknownExtension: string;
  end;

var
  LogQueue: TThreadList<string> = nil;

implementation

function IntToHexStr(Value: Integer; Digits: Integer): string;
begin
  result := '0x' + IntToHex(Value, Digits);
end;

function GetHexStr(S: array of byte; Size: integer): string;
var
  i: integer;
begin
  result := '';
  if Size > 0 then begin
    for i := 0 to Size - 1 do
      result := result + IntToHexStr(S[i], 2) + ' ';
    result := Trim(result);
  end;
end;


{ THeader }

function THeader.GlobalColorTableFlag: boolean;             // 1000 0000 0x80
begin
  result := ((TheBits and $80) shr 7) = 1;
end;

function THeader.ColorResolution: integer;
begin
  result := ((TheBits and $70) shr 4) + 1;
end;

function THeader.SortFlag: boolean;
begin
  result := ((TheBits and $8) shr 3) = 1;
end;

function THeader.SizeOfGlobalColorTable: integer;
begin
  result := (TheBits and $7) + 1;
end;

function THeader.SizeOfColorTable: integer;
begin
  if GlobalColorTableFlag then
    result := Round(Power(2, SizeOfGlobalColorTable)) * 3
  else
    result := 0;
end;

{ TGraphicControlExtension }

function TGraphicControlExtension.DisposalMothod: integer;  // 0001 1100 0x1C
begin
  result := (TheBits and $2C) shr 2;
end;

function TGraphicControlExtension.UserInputFlag: boolean;   // 0000 0010  0x2
begin
  result := ((TheBits and $2) shr 1) = 1;
end;

function TGraphicControlExtension.TransparentColorFlag: boolean;
begin
  result := (TheBits and $1) = 1;
end;

{ TImageBlock }

function TImageBlock.LocalColorTableFlag: boolean;          // 1000 0000 0x80
begin
  result := ((TheBits and $80) shr 7) = 1;
end;

function TImageBlock.InterlaceFlag: boolean;                // 0100 0000 0x40
begin
  result := ((TheBits and $40) shr 6) = 1;
end;

function TImageBlock.SortFlag: boolean;                     // 0010 0000 0x20
begin
  result := ((TheBits and $20) shr 5) = 1;
end;

function TImageBlock.SizeOfLocalColorTable: integer;        // 0000 0111 0x7
begin
  result := TheBits and $7 + 1;
end;

function TImageBlock.SizeOfColorTable: integer;
begin
  if LocalColorTableFlag then
    result := Round(Power(2, SizeOfLocalColorTable)) * 3
  else
    result := 0;
end;

{ TAnimatedGifClass }

constructor TAnimatedGifClass.Create;
begin
  MS := TMemoryStream.Create;
  FBlockList := TList.Create;
  FTrailer := 0;
  FProcessing := false;
end;

destructor TAnimatedGifClass.Destroy;
begin
  try
    FreeAndNil(MS);

    DataClear;
    FreeAndNil(FBlockList);
  except
    on e: Exception do
      if Assigned(LogQueue) then
        LogQueue.Add('TAnimatedGifClass.Destroy: ' + e.Message);
  end;

  inherited;
end;

function TAnimatedGifClass.GetBlockListCount: integer;
begin
  result := FBlockList.Count;
end;

function TAnimatedGifClass.GetBlockListRec(index: integer): PItemRec;
begin
  if (index >= 0) and (index < FBlockList.Count) then
    result := FBlockList.Items[index]
  else
    result := nil;
end;

// Clear

procedure TAnimatedGifClass.DataClear;
var
  i: integer;
  Rec: PItemRec;
begin
  if FBlockList.Count > 0 then begin
    try
      for i := 0 to FBlockList.Count - 1 do begin
        Rec := FBlockList.Items[i];
        if Rec <> nil then begin
          case Rec^.BlockType of
            btHeader: Dispose(PHeader(Rec^.Block));
            btAppExt: ApplicationExtensionClear(Rec^.Block);
            btCtrlExt: Dispose(PGraphicControlExtension(Rec^.Block));
            btImgBlock: ImageBlockClear(Rec^.Block);
            btCommentExt: CommentExtensionClear(Rec^.Block);
            btTextExt: TextExtensionClear(Rec^.Block);
            // btTrailer:
            // btUnknown:
          end;
          Dispose(Rec);
        end;
      end;
    except
      on e: Exception do
        if Assigned(LogQueue) then
          LogQueue.Add('TAnimatedGifClass.DataClear: ' + e.Message);
    end;
    FBlockList.Clear;
  end;
end;

procedure TAnimatedGifClass.ApplicationExtensionClear(Rec: PApplicationExtension);
var
  i: integer;
  BRec: PBlock;
begin
  try
    if Rec = nil then exit;
    if Rec^.Blocks <> nil then begin
      if Rec^.Blocks.Count > 0 then begin
        for i := 0 to Rec^.Blocks.Count - 1 do begin
          BRec:= Rec^.Blocks.Items[i];
          if BRec <> nil then
            Dispose(BRec);
        end;
      end;
      Rec^.Blocks.Free;
    end;
    Dispose(Rec);
  except
    on e: Exception do
      if Assigned(LogQueue) then
        LogQueue.Add('TAnimatedGifClass.ApplicationExtensionClear: ' + e.Message);
  end;
end;

procedure TAnimatedGifClass.ImageBlockClear(Rec: PImageBlock);
var
  i: integer;
  BRec: PBlock;
begin
  try
    if Rec = nil then exit;
    if Rec^.Blocks <> nil then begin
      if Rec^.Blocks.Count > 0 then begin
        for i := 0 to Rec^.Blocks.Count - 1 do begin
          BRec := Rec^.Blocks.Items[i];
          if BRec <> nil then
              Dispose(BRec);
        end;
      end;
      Rec^.Blocks.Free;
    end;
    Dispose(Rec);
  except
    on e: Exception do
      if Assigned(LogQueue) then
        LogQueue.Add('TAnimatedGifClass.ImageBlockClear: ' + e.Message);
  end;
end;

procedure TAnimatedGifClass.CommentExtensionClear(Rec: PCommentExtension);
var
  i: integer;
  BRec: PBlock;
begin
  try
    if Rec = nil then exit;
    if Rec^.Blocks <> nil then begin
      if Rec^.Blocks.Count > 0 then begin
        for i := 0 to Rec^.Blocks.Count - 1 do begin
          BRec := Rec^.Blocks.Items[i];
          if BRec <> nil then
            Dispose(BRec);
        end;
      end;
      Rec^.Blocks.Free;
    end;
    Dispose(Rec);
  except
    on e: Exception do
      if Assigned(LogQueue) then
        LogQueue.Add('TAnimatedGifClass.CommentExtensionClear: ' + e.Message);
  end;
end;

procedure TAnimatedGifClass.TextExtensionClear(Rec: PTextExtension);
var
  i: integer;
  BRec: PBlock;
begin
  try
    if Rec = nil then exit;
    if Rec^.Blocks <> nil then begin
      if Rec^.Blocks.Count > 0 then begin
        for i := 0 to Rec^.Blocks.Count - 1 do begin
          BRec := Rec^.Blocks.Items[i];
          if BRec <> nil then
            Dispose(BRec);
        end;
      end;
      Rec^.Blocks.Free;
    end;
    Dispose(Rec);
  except
    on e: Exception do
      if Assigned(LogQueue) then
        LogQueue.Add('TAnimatedGifClass.TextExtensionClear: ' + e.Message);
  end;
end;

procedure TAnimatedGifClass.LoadData(Filename: string);
begin
  if (Filename <> '') and (FileExists(Filename)) then begin
    MS.Clear;
    MS.Seek(0, TSeekOrigin.soBeginning);
    MS.LoadFromFile(Filename);
    DecodeNow;
  end;
end;

// decode

procedure TAnimatedGifClass.DecodeNow;
var
  Rec: PItemRec;
  B: Byte;
  UnKnownExtensionCount: integer;
begin
  try
    if MS.Size > 0 then begin
      FControlExtensionCount := 0;
      FImageBlockCount := 0;
      MS.Seek(0, TSeekOrigin.soBeginning);

      DecodeHeader;
      MS.Read(FNextByte, 1);
      UnKnownExtensionCount := 0;
      while (FNextByte <> $3B) and (MS.Position <= MS.Size) do begin       // Trailer
        if FNextByte = $2C then
          DecodeImageBlock
        else if FNextByte = $21 then begin
          MS.Read(FNextByte, 1);
          if FNextByte = $F9 then
            DecodeGraphicControlExtension
          else if FNextByte = $FE then
            DecodeCommentExtension
          else if FNextByte = $01 then
            DecodePlainTextExtension
          else if FNextByte = $FF then begin
            DecodeApplicationExtension;
          end;
        end else begin
          DecodeUnknownExtension;
          Inc(UnKnownExtensionCount);
          if UnKnownExtensionCount > 8 then begin
            // ShowMessage('Too many unknown extensions !');
            exit;
          end;
        end;
        MS.Read(FNextByte, 1);
      end;

      // temp
      if MS.Position < MS.Size then
        MS.Read(B, 1);

      if FNextByte = $3B then begin
        FTrailer := FNextByte;
        New(Rec);
        Rec^.BlockType := btTrailer;
        Rec^.Caption := 'Trailer';
        Rec^.Info := '';
        Rec^.Block := nil;
        FBlockList.Add(Rec);
      end;
    end;
  except
    on e: Exception do
      if Assigned(LogQueue) then
        LogQueue.Add('TAnimatedGifClass.DecodeNow: ' + e.Message);
  end;
end;

// Decoder

procedure TAnimatedGifClass.DecodeHeader;
var
  Rec: PHeader;
  LVRec: PItemRec;
begin
  try
    New(Rec);
    MS.Read(Rec^.Signature, 3);
    MS.Read(Rec^.Version, 3);
    MS.Read(Rec^.LogicalScreenWidth, 2);
    MS.Read(Rec^.LogicalScreenHeight, 2);
    MS.Read(Rec^.TheBits, 1);
    MS.Read(Rec^.BackgroundColorIndex, 1);
    MS.Read(Rec^.PixelAspectRatio, 1);
    if Rec^.GlobalColorTableFlag then
      MS.Read(Rec^.ColorTable, Rec^.SizeOfColorTable);

    New(LVRec);
    LVRec^.BlockType := btHeader;
    LVRec^.Caption := 'Header';
    LVRec^.Info := '';
    LVRec^.Block := Rec;
    FBlockList.Add(LVRec);
  except
    on e: Exception do
      if Assigned(LogQueue) then
        LogQueue.Add('TAnimatedGifClass.DecodeHeader: ' + e.Message);
  end;
end;

procedure TAnimatedGifClass.DecodeGraphicControlExtension;
var
  Rec: PGraphicControlExtension;
  LVRec: PItemRec;
begin
  try
    New(Rec);
    Rec^.Address := MS.Position - 2;
    Rec^.Index := FControlExtensionCount;
    Inc(FControlExtensionCount);
    Rec^.ExtensionIntroducer := $21;
    Rec^.GraphicControlLabel := $F9;
    MS.Read(Rec^.BlockSize, 1);
    MS.Read(Rec^.TheBits, 1);
    MS.Read(Rec^.DelayTime, 2);
    MS.Read(Rec^.TransparentColorIndex, 1);
    MS.Read(Rec^.BlockTerminator, 1);

    Rec^.Size := MS.Position - 2 - Rec^.Address;

    New(LVRec);
    LVRec^.BlockType := btCtrlExt;
    LVRec^.Caption := 'GraphicControlExtension(' + IntToStr(Rec^.Index) + ')';
    LVRec^.Info := '';
    LVRec^.Block := Rec;
    FBlockList.Add(LVRec);
  except
    on e: Exception do
      if Assigned(LogQueue) then
        LogQueue.Add('TAnimatedGifClass.DecodeGraphicControlExtension: ' + e.Message);
  end;
end;

procedure TAnimatedGifClass.DecodeImageBlock;
var
  Rec: PImageBlock;
  BRec: PBlock;
  LVRec: PItemRec;
begin
  try
    New(Rec);
    Rec^.Address := MS.Position - 1;
    Rec^.Index := FImageBlockCount;
    Inc(FImageBlockCount);
    Rec^.ImageSeparator := $2C;
    MS.Read(Rec^.ImageLeftPosition, 2);
    MS.Read(Rec^.ImageTopPosition, 2);
    MS.Read(Rec^.ImageWidth, 2);
    MS.Read(Rec^.ImageHeight, 2);
    MS.Read(Rec^.TheBits, 1);
    if Rec^.LocalColorTableFlag then
      MS.Read(Rec^.LocalColorTable, Rec^.SizeOfColorTable);
    MS.Read(Rec^.LZWMinimumCodeSide, 1);

    Rec^.Blocks := TList{<PBlock>}.Create;
    repeat
      New(BRec);
      Rec^.Blocks.Add(BRec);
      BRec^.BlockSize := 0;
      MS.Read(BRec^.BlockSize, 1);
      if BRec^.BlockSize > 0 then
        MS.Read(BRec^.Data, BRec^.BlockSize);
    until BRec^.BlockSize = 0;

    Rec^.Size := MS.Position - 1 - Rec^.Address;

    New(LVRec);
    LVRec^.BlockType := btImgBlock;
    LVRec^.Caption := 'ImageBlock(' + IntToStr(Rec^.Index) + ')';
    LVRec^.Info := '';
    LVRec^.Block := Rec;
    FBlockList.Add(LVRec);
  except
    on e: Exception do
      if Assigned(LogQueue) then
        LogQueue.Add('TAnimatedGifClass.DecodeImageBlock: ' + e.Message);
  end;
end;

procedure TAnimatedGifClass.DecodeCommentExtension;
var
  Rec: PCommentExtension;
  BRec: PBlock;
  LVRec: PItemRec;
begin
  try
    New(Rec);
    Rec^.Address := MS.Position - 1;
    Rec^.ExtensionIntroducer := $21;
    Rec^.CommentLabel := $FE;

    Rec^.Blocks := TList{<PBlock>}.Create;
    repeat
      New(BRec);
      Rec^.Blocks.Add(BRec);
      BRec^.BlockSize := 0;
      MS.Read(BRec^.BlockSize, 1);
      if BRec^.BlockSize > 0 then
        MS.Read(BRec^.Data, BRec^.BlockSize);
    until BRec^.BlockSize = 0;

    Rec^.Size := MS.Position - 1 - Rec^.Address;

    New(LVRec);
    LVRec^.BlockType := btCommentExt;
    LVRec^.Caption := 'CommentExtension';
    LVRec^.Info := '';
    LVRec^.Block := Rec;
    FBlockList.Add(LVRec);
  except
    on e: Exception do
      if Assigned(LogQueue) then
        LogQueue.Add('TAnimatedGifClass.DecodeCommentExtension: ' + e.Message);
  end;
end;

procedure TAnimatedGifClass.DecodePlainTextExtension;
var
  Rec: PTextExtension;
  BRec: PBlock;
  LVRec: PItemRec;
begin
 try
    New(Rec);
    Rec^.Address := MS.Position - 1;
    Rec^.ExtensionIntroducer := $21;
    Rec^.PlainTextLabel := $01;
    MS.Read(Rec^.BlockSize, 1);  // $0C
    MS.Read(Rec^.TextGridLeftPosition, 2);
    MS.Read(Rec^.TextGridTopPosition, 2);
    MS.Read(Rec^.TextGridWidth, 2);
    MS.Read(Rec^.TextGridHeight, 2);
    MS.Read(Rec^.CharacterCellWidth, 1);
    MS.Read(Rec^.CharacterCellHeight, 1);
    MS.Read(Rec^.TextForegroundColorIndex, 1);
    MS.Read(Rec^.TextBackgroundColorIndex, 1);

    Rec^.Blocks := TList{<PBlock>}.Create;
    repeat
      New(BRec);
      Rec^.Blocks.Add(BRec);
      BRec^.BlockSize := 0;
      MS.Read(BRec^.BlockSize, 1);
      if BRec^.BlockSize > 0 then
        MS.Read(BRec^.Data, BRec^.BlockSize);
    until BRec^.BlockSize = 0;

    Rec^.Size := MS.Position - 1 - Rec^.Address;

    New(LVRec);
    LVRec^.BlockType := btTextExt;
    LVRec^.Caption := 'TextExtension';
    LVRec^.Info := '';
    LVRec^.Block := Rec;
    FBlockList.Add(LVRec);
  except
    on e: Exception do
      if Assigned(LogQueue) then
        LogQueue.Add('TAnimatedGifClass.DecodePlainTextExtension: ' + e.Message);
  end;
end;

procedure TAnimatedGifClass.DecodeApplicationExtension;
var
  Rec: PApplicationExtension;
  BRec: PBlock;
  LVRec: PItemRec;
begin
  try
    New(Rec);
    Rec^.ExtensionIntroducer := $21;
    Rec^.ExtensionLabel := $FF;
    Rec^.Address := MS.Position - 2;
    MS.Read(Rec^.BlockSize, 1);
    MS.Read(Rec^.ApplicationIdentifier, 8);
    MS.Read(Rec^.ApplicationAuthenticationCode, 3);

    Rec^.Blocks := TList{<PBlock>}.Create;
    repeat
      New(BRec);
      Rec^.Blocks.Add(BRec);
      BRec^.BlockSize := 0;
      MS.Read(BRec^.BlockSize, 1);
      if BRec^.BlockSize > 0 then
        MS.Read(BRec^.Data, BRec^.BlockSize);
    until BRec^.BlockSize = 0;

    Rec^.Size := MS.Position - 1 - Rec^.Address;

    New(LVRec);
    LVRec^.BlockType := btAppExt;
    LVRec^.Caption := 'AppExtension';;
    LVRec^.Info := '';
    LVRec^.Block := Rec;
    FBlockList.Add(LVRec);
  except
    on e: Exception do
      if Assigned(LogQueue) then
        LogQueue.Add('TAnimatedGifClass.DecodeApplicationExtension: ' + e.Message);
  end;
end;

procedure TAnimatedGifClass.DecodeUnknownExtension;
var
  LVRec: PItemRec;
begin
  try
    New(LVRec);
    LVRec^.BlockType := btUnknown;
    LVRec^.Caption := 'Unknown Extension';
    LVRec^.Info := ''; //'"' + IntToHexStr(FNextByte, 2) + '" in ' + IntToHexStr(MS.Position - 1, 8);
    LVRec^.Block := nil;
    FBlockList.Add(LVRec);
  except
    on e: Exception do
      if Assigned(LogQueue) then
        LogQueue.Add('TAnimatedGifClass.DecodeUnknownExtension: ' + e.Message);
  end;
end;

// Encode

procedure TAnimatedGifClass.SaveData(Filename: string);
begin
  if EncodeNow then
    MS.SaveToFile(FileName);
end;

procedure TAnimatedGifClass.WriteData(Rec: PBlock);
const
  Terminater: byte = 0;
begin
  try
    if Rec^.BlockSize > 0 then begin
      MS.Write(Rec^.BlockSize, 1);
      MS.Write(Rec^.Data, Rec^.BlockSize);
    end else
      MS.Write(Terminater, 1);
  except

  end;
end;

function TAnimatedGifClass.EncodeNow: boolean;
var
  i: integer;
  Rec: PItemRec;
begin
  result := false;

  try
    if FBlockList.Count > 0 then begin
      MS.Clear;
      MS.Seek(0, TSeekOrigin.soBeginning);
      for i := 0 to FBlockList.Count -1 do begin
        Rec := FBlockList.Items[i];
        if Rec <> nil then begin
          case Rec^.BlockType of
            btHeader: EncodeHeader(Rec^.Block);
            btAppExt: EncodeApplicationExtension(Rec^.Block);
            btCtrlExt: EncodeGraphicControlExtension(Rec^.Block);
            btImgBlock: EncodeImageBlock(Rec^.Block);
            btCommentExt: EncodeCommentExtension(Rec^.Block);
            btTextExt: EncodePlainTextExtension(Rec^.Block);
            btTrailer: MS.Write(FTrailer, 1)
            // btUnknown: ShowUnknownExtension;
          end;
        end;
      end;
      result := true;
    end;
  except
    on e: Exception do
      if Assigned(LogQueue) then
        LogQueue.Add('TAnimatedGifClass.EncodeNow: ' + e.Message);
  end;
end;


procedure TAnimatedGifClass.EncodeHeader(Rec: PHeader);
begin
  if Rec <> nil then begin
    try
      MS.Write(Rec^.Signature, 3);
      MS.Write(Rec^.Version, 3);
      MS.Write(Rec^.LogicalScreenWidth, 2);
      MS.Write(Rec^.LogicalScreenHeight, 2);
      MS.Write(Rec^.TheBits, 1);
      MS.Write(Rec^.BackgroundColorIndex, 1);
      MS.Write(Rec^.PixelAspectRatio, 1);
      if Rec^.GlobalColorTableFlag then
        MS.Write(Rec^.ColorTable, Rec^.SizeOfColorTable);
    except
      on e: Exception do
        if Assigned(LogQueue) then
          LogQueue.Add('TAnimatedGifClass.EncodeHeader: ' + e.Message);
    end;
  end;
end;

procedure TAnimatedGifClass.EncodeGraphicControlExtension(Rec: PGraphicControlExtension);
begin
  if Rec <> nil then begin
    try
      MS.Write(Rec^.ExtensionIntroducer, 1);
      MS.Write(Rec^.GraphicControlLabel, 1);
      MS.Write(Rec^.BlockSize, 1);
      MS.Write(Rec^.TheBits, 1);
      MS.Write(Rec^.DelayTime, 2);
      MS.Write(Rec^.TransparentColorIndex, 1);
      MS.Write(Rec^.BlockTerminator, 1);
    except
      on e: Exception do
        if Assigned(LogQueue) then
          LogQueue.Add('TAnimatedGifClass.EncodeHeader: ' + e.Message);
    end;
  end;
end;

procedure TAnimatedGifClass.EncodeImageBlock(Rec: PImageBlock);
var
  i: integer;
begin
  if Rec <> nil then begin
    try
      MS.Write(Rec^.ImageSeparator, 1);;
      MS.Write(Rec^.ImageLeftPosition, 2);
      MS.Write(Rec^.ImageTopPosition, 2);
      MS.Write(Rec^.ImageWidth, 2);
      MS.Write(Rec^.ImageHeight, 2);
      MS.Write(Rec^.TheBits, 1);
      if Rec^.LocalColorTableFlag then
        MS.Write(Rec^.LocalColorTable, Rec^.SizeOfColorTable);
      MS.Write(Rec^.LZWMinimumCodeSide, 1);

      if (Rec^.Blocks <> nil) and (Rec^.Blocks.Count > 0) then
        for i := 0 to Rec^.Blocks.Count - 1 do
          WriteData(Rec^.Blocks.Items[i]);
    except
      on e: Exception do
        if Assigned(LogQueue) then
          LogQueue.Add('TAnimatedGifClass.EncodeImageBlock: ' + e.Message);
    end;
  end;
end;

procedure TAnimatedGifClass.EncodeCommentExtension(Rec: PCommentExtension);
var
  i: integer;
begin
  if Rec <> nil then begin
    try
      MS.Write(Rec^.ExtensionIntroducer, 1);
      MS.Write(Rec^.CommentLabel, 1);

      if (Rec^.Blocks <> nil) and (Rec^.Blocks.Count > 0) then
        for i := 0 to Rec^.Blocks.Count - 1 do
          WriteData(Rec^.Blocks.Items[i]);
    except
      on e: Exception do
        if Assigned(LogQueue) then
          LogQueue.Add('TAnimatedGifClass.EncodeCommentExtension: ' + e.Message);
    end;
  end;

end;

procedure TAnimatedGifClass.EncodePlainTextExtension(Rec: PTextExtension);
var
  i: integer;
begin
  if Rec <> nil then begin
    try
      MS.Write(Rec^.ExtensionIntroducer, 1);
      MS.Write(Rec^.PlainTextLabel, 1);
      MS.Write(Rec^.BlockSize, 1);
      MS.Write(Rec^.TextGridLeftPosition, 2);
      MS.Write(Rec^.TextGridTopPosition, 2);
      MS.Write(Rec^.TextGridWidth, 2);
      MS.Write(Rec^.TextGridHeight, 2);
      MS.Write(Rec^.CharacterCellWidth, 1);
      MS.Write(Rec^.CharacterCellHeight, 1);
      MS.Write(Rec^.TextForegroundColorIndex, 1);
      MS.Write(Rec^.TextBackgroundColorIndex, 1);

      if (Rec^.Blocks <> nil) and (Rec^.Blocks.Count > 0) then
        for i := 0 to Rec^.Blocks.Count - 1 do
          WriteData(Rec^.Blocks.Items[i]);
    except
      on e: Exception do
        if Assigned(LogQueue) then
          LogQueue.Add('TAnimatedGifClass.EncodePlainTextExtension: ' + e.Message);
    end;
  end;

end;

procedure TAnimatedGifClass.EncodeApplicationExtension(Rec: PApplicationExtension);
var
  i: integer;
begin
  if Rec <> nil then begin
    try
      MS.Write(Rec^.ExtensionIntroducer, 1);
      MS.Write(Rec^.ExtensionLabel, 1);
      MS.Write(Rec^.BlockSize, 1);
      MS.Write(Rec^.ApplicationIdentifier, 8);
      MS.Write(Rec^.ApplicationAuthenticationCode, 3);

      if (Rec^.Blocks <> nil) and (Rec^.Blocks.Count > 0) then
        for i := 0 to Rec^.Blocks.Count - 1 do
          WriteData(Rec^.Blocks.Items[i]);
    except
      on e: Exception do
        if Assigned(LogQueue) then
          LogQueue.Add('TAnimatedGifClass.EncodeApplicationExtension: ' + e.Message);
    end;
  end;

end;

// Gif info

function TAnimatedGifClass.GetHeaderInfo(Rec: PHeader): string;
var
  i: integer;
begin
  result := '';

  result := result + 'Signature: ' + UnicodeString(Rec^.Signature) + #13#10;
  result := result + 'Version: ' + UnicodeString(Rec^.Version) + #13#10;
  result := result + 'Logical Screen Width: ' + IntToStr(Rec^.LogicalScreenWidth) + #13#10;
  result := result + 'Logical Screen Height: ' + IntToStr(Rec^.LogicalScreenHeight) + #13#10;
  result := result + 'TheBits: ' + IntToHexStr(Rec^.TheBits, 2) + #13#10;
  result := result + 'Global Color Table Flag: ' + BoolToStr(Rec^.GlobalColorTableFlag, true) + #13#10;
  result := result + 'Color Resolution: ' + IntToStr(Rec^.ColorResolution) + #13#10;
  result := result + 'Sort Flag: ' + BoolToStr(Rec^.SortFlag, true) + #13#10;
  result := result + 'Size of Global Color Table: ' + IntToStr(Rec^.SizeOfGlobalColorTable) + #13#10;
  result := result + 'Background Color Index: ' + IntToStr(Rec^.BackgroundColorIndex) + #13#10;
  result := result + 'Pixel Aspect Ratio: ' + IntToStr(Rec^.PixelAspectRatio) + #13#10;
  result := result + 'Color Table Size: ' + IntToStr(Rec^.SizeOfColorTable) + #13#10;
  if Rec^.SizeOfColorTable > 0 then
    for i := 0 to  + Rec^.SizeOfColorTable - 1 do begin
      result := result + IntToHexStr(Rec^.ColorTable[i], 2);
      if i mod 3 = 2 then
      result := result + '-';
    end;
end;

function TAnimatedGifClass.GetGraphicControlExtensionInfo(Rec: PGraphicControlExtension): string;
begin
  result := '';

  result := result + 'Control Extension Address: ' + IntToHexStr(Rec^.Address, 8) + #13#10;
  result := result + 'Control Extension Size: ' + IntToHexStr(Rec^.Size, 8) + #13#10;
  result := result + 'Control Extension Index: ' + IntToStr(Rec^.Index) + #13#10;
  result := result + 'Extension Introducer: ' + IntToHexStr(Rec^.ExtensionIntroducer, 2) + #13#10;
  result := result + 'Graphic Control Label: ' + IntToHexStr(Rec^.GraphicControlLabel, 2) + #13#10;
  result := result + 'Block Size: ' + IntToStr(Rec^.BlockSize) + #13#10;
  result := result + 'Disposal Mothod: ' + IntToStr(Rec^.DisposalMothod) + #13#10;
  result := result + 'User Input Flag: ' + BoolToStr(Rec^.UserInputFlag, true) + #13#10;
  result := result + 'Transparent Color Flag: ' + BoolToStr(Rec^.TransparentColorFlag, true) + #13#10;
  result := result + 'Delay Time: ' + IntToStr(Rec^.DelayTime) + #13#10;
  result := result + 'Transparent Color Index: ' + IntToStr(Rec^.TransparentColorIndex) + #13#10;
  result := result + 'Block Terminator: ' + IntToStr(Rec^.BlockTerminator);

end;

function TAnimatedGifClass.GetImageBlockInfo(Rec: PImageBlock): string;
var
  i: integer;
  BRec: PBlock;
  DataSize: integer;
  Count: integer;
  c0, c1: integer;
begin
  result := '';

  result := result + 'Image Block Address: ' + IntToHexStr(Rec^.Address, 8) + #13#10;
  result := result + 'Image Block Size: ' + IntToHexStr(Rec^.Size, 8) + #13#10;
  result := result + 'Image Block Index: ' + IntToStr(Rec^.Index) + #13#10;
  result := result + 'Image Separator: ' + IntToHexStr(Rec^.ImageSeparator, 2) + #13#10;
  result := result + 'Image Left Position: ' + IntToStr(Rec^.ImageLeftPosition) + #13#10;
  result := result + 'Image Top Position: ' + IntToStr(Rec^.ImageTopPosition) + #13#10;
  result := result + 'Image Width: ' + IntToStr(Rec^.ImageWidth) + #13#10;
  result := result + 'Image Height: ' + IntToStr(Rec^.ImageHeight) + #13#10;
  result := result + 'Local Color Table Flag: ' + BoolToStr(Rec^.LocalColorTableFlag, true) + #13#10;
  result := result + 'Interlace Flag: ' + BoolToStr(Rec^.InterlaceFlag, true) + #13#10;
  result := result + 'Sort Flag: ' + BoolToStr(Rec^.SortFlag, true) + #13#10;
  result := result + 'Size Of Local Color Table: ' + IntToStr(Rec^.SizeOfLocalColorTable) + #13#10;
  result := result + 'Color Table Size(real): ' + IntToStr(Rec^.SizeOfColorTable) + #13#10;
  result := result + 'LZW Minimum Code Side: ' + IntToStr(Rec^.LZWMinimumCodeSide) + #13#10;

  if Rec^.LocalColorTableFlag then begin
    result := result + 'Local color table' + #13#10;
    c0 := 0;
    c1 := 0;
    for i := 0 to Rec^.SizeOfColorTable - 1 do begin
      result := result + IntToHexStr(Rec^.LocalColorTable[i], 2);
      inc(c0);
      if c0 = 3 then begin
        c0 := 0;
        Inc(c1);
        if c1 >= 4 then begin
          c1 := 0;
          result := result + #13#10;
        end else
          result := result + ' ';
      end;
    end;
  end;

  result := result + #13#10;

  Count := Rec^.Blocks.Count - 1;
  if Count < 0 then
    Count := 0;
  result := result + 'Data Block Count: ' + IntToStr(Count) + #13#10;

  DataSize := 0;
  if (Rec^.Blocks <> nil) and (Rec^.Blocks.Count > 0) then begin
    for i := 0 to Rec^.Blocks.Count - 1 do begin
      BRec := Rec^.Blocks.Items[i];
      DataSize := DataSize + BRec^.BlockSize;
    end;
  end;
  result := result + 'Data Length: ' + IntToStr(DataSize);
end;

function TAnimatedGifClass.GetCommentExtensionInfo(Rec: PCommentExtension): string;
var
  i: integer;
  BRec: PBlock;
  S: string;
begin
  result := '';

  try
    result := result + 'Comment Extension Address: ' + IntToHexStr(Rec^.Address, 8) + #13#10;
    result := result + 'Comment Extension Size: ' + IntToHexStr(Rec^.Size, 8) + #13#10;
    result := result + 'Extension Introducer: ' + IntToHexStr(Rec^.ExtensionIntroducer, 2) + #13#10;
    result := result + 'Comment Label: ' + IntToHexStr(Rec^.CommentLabel, 2) + #13#10;

    if Rec^.Blocks.Count > 0 then
      for i := 0 to Rec^.Blocks.Count - 1 do begin
        BRec := Rec^.Blocks.Items[i];
        result := result + 'Block(' + IntToStr(i) + ') Size: ' + IntToStr(BRec^.BlockSize) + #13#10;
        SetString(S, PAnsiChar(@BRec^.Data[0]), BRec^.BlockSize);
        result := result + 'Block(' + IntToStr(i) + ') Content: ' + S + #13#10;
      end;
  except
    on e: Exception do
      if Assigned(LogQueue) then
        LogQueue.Add('TAnimatedGifClass.ShowCommentExtensionInfo: ' + e.Message);
  end;
end;

function TAnimatedGifClass.GetPlainTextExtensionInfo(Rec: PTextExtension): string;
var
  i: integer;
  BRec: PBlock;
  S: string;
begin
  result := '';

  try
    result := result + 'Text Extension Address: ' + IntToHexStr(Rec^.Address, 8) + #13#10;
    result := result + 'Text Extension Size: ' + IntToHexStr(Rec^.Size, 8) + #13#10;
    result := result + 'Extension Introducer: ' + IntToHexStr(Rec^.ExtensionIntroducer, 2) + #13#10;
    result := result + 'Plain Text Label: ' + IntToHexStr(Rec^.PlainTextLabel, 2) + #13#10;
    result := result + 'BlockSize: ' + IntToHexStr(Rec^.BlockSize, 2) + #13#10;
    result := result + 'Text Grid Left Position: ' + IntToStr(Rec^.TextGridLeftPosition) + #13#10;
    result := result + 'Text Grid Top Position: ' + IntToStr(Rec^.TextGridTopPosition) + #13#10;
    result := result + 'Text Grid Width: ' + IntToStr(Rec^.TextGridWidth) + #13#10;
    result := result + 'Text Grid Height: ' + IntToStr(Rec^.TextGridHeight) + #13#10;
    result := result + 'Character Cell Width: ' + IntToStr(Rec^.CharacterCellWidth) + #13#10;
    result := result + 'CharacterCellHeight: ' + IntToStr(Rec^.CharacterCellHeight) + #13#10;
    result := result + 'Text Foreground Color Index' + IntToStr(Rec^.TextForegroundColorIndex) + #13#10;
    result := result + 'Text Background Color Index: ' + IntToStr(Rec^.TextBackgroundColorIndex) + #13#10;

    if Rec^.Blocks.Count > 0 then
      for i := 0 to Rec^.Blocks.Count - 1 do begin
        BRec := Rec^.Blocks.Items[i];
        result := result + 'Block(' + IntToStr(i) + ') Size: ' + IntToStr(BRec^.BlockSize) + #13#10;
        SetString(S, PAnsiChar(@BRec^.Data[0]), BRec^.BlockSize);
        result := result + 'Block(' + IntToStr(i) + ') Content: ' + S + #13#10;
      end;
  except
    on e: Exception do
      if Assigned(LogQueue) then
        LogQueue.Add('TAnimatedGifClass.ShowPlainTextExtensionInfo: ' + e.Message);
  end;
end;

function TAnimatedGifClass.GetApplicationExtensionInfo(Rec: PApplicationExtension): string;
var
  BRec: PBlock;
  i: integer;
begin
  result := '';

  try
    result := result + 'Application Extension Address: ' + IntToHexStr(Rec^.Address, 8) + #13#10;
    result := result + 'Application Extension Size: ' + IntToHexStr(Rec^.Size, 8) + #13#10;
    result := result + 'ExtensionIntroducer: ' + IntToHexStr(Rec^.ExtensionIntroducer, 2) + #13#10;
    result := result + 'ExtensionLabel: ' + IntToHexStr(Rec^.ExtensionLabel, 2) + #13#10;
    result := result + 'BlockSize: ' + IntToHexStr(Rec^.BlockSize, 2) + #13#10;
    result := result + 'ApplicationIdentifier: ' + UnicodeString(Rec^.ApplicationIdentifier) + #13#10;
    result := result + 'ApplicationAuthenticationCode: ' + UnicodeString(Rec^.ApplicationAuthenticationCode) + #13#10;

    if Rec^.Blocks.Count > 0 then
      for i := 0 to Rec^.Blocks.Count - 1 do begin
        BRec := Rec^.Blocks.Items[i];
        result := result + 'Block(' + IntToStr(i) + ') Size: ' + IntToStr(BRec^.BlockSize) + #13#10;
        result := result + 'Block(' + IntToStr(i) + ') Content: ' + GetHexStr(BRec^.Data, BRec^.BlockSize) + #13#10;
      end;
  except
    on e: Exception do
      if Assigned(LogQueue) then
        LogQueue.Add('TAnimatedGifClass.ShowApplicationExtensionInfo: ' + e.Message);
  end;
end;

function TAnimatedGifClass.GetUnknownExtension: string;
begin
  result := '';
end;



end.
